#ifndef DVXML_DOC_H
#define DVXML_DOC_H

#include <stdexcept>
#include <iterator>
#include <dvutil/tostring.h>
#include <dvutil/file.h>
#include <xmlwrapp/xmlwrapp.h>

#include <dvxml/exception.h>
#include <dvxml/noconst.h>
#include <dvxml/node.h>

namespace Dv {
namespace Xml {

/** Convenience interface to xml::document.
 *
 */
class Document: public xml::document {
public:
  /** Default ctor. */
  explicit Document();

  /** Construct a Dv::Xml::Document out of a (copy of a) Dv::Xml::Node object.
   * @param ref Dv::Xml::Node::Ref, a copy of which will be used as the root
   *   of the new Dv::Xml::Document object.
   * @exception Dv::Xml::Exception if node is not valid.
   */
  explicit Document(const Dv::Xml::Node::Ref& ref) throw (Dv::Xml::Exception) ;

  /** Parse a Dv::Xml::Document from a file
   * @param file containing document
   * @exception Dv::Xml::Exception upon any error
   */
  explicit Document(const Dv::Util::File& file) throw (Dv::Xml::Exception) ;

  /** Create a Dv::Xml::Document with a root labeled with given name and content. 
   * @param name to be used as the label of the root of the document
   * @param content to be used as the content of the root of the document
   * @exception Dv::Xml::Exception upon any error
   */
  explicit Document(const std::string& name, const std::string& content)
    throw (Dv::Xml::Exception);

  /** Create a Dv::Xml::Document with a root labeled with given name.
   * @param name to be used as the label of the root of the document
   * @exception Dv::Xml::Exception upon any error
   */
  explicit Document(const std::string& name) throw (Dv::Xml::Exception);

  /** Parse a Dv::Xml::Document from an array of characters .
   * @param pc pointer to the start of the data
   * @param size of the data array
   * @exception Dv::Xml::Exception upon any error
   */
  explicit Document(const char* pc, size_t size) throw (Dv::Xml::Exception) ;

  /** Create a Dv::Xml::Document by applying an XSLT stylesheet to
   * another Document.
   * @param stylesheet filename of xslt file
   * @param doc input document
   * @exception Dv::Xml::Exception upon any error
   */
  explicit Document(const std::string& stylesheet, Dv::Xml::Document& doc)
    throw (Dv::Xml::Exception) ;

  /** Destructor. */
  virtual ~Document();

  /** Parse a Dv::Xml::Document from an array of characters . The
   * result replaces the current contents of this document.
   * @param data pointer to the start of the data
   * @param size of the data array
   * @return reference to this document
   * @exception Dv::Xml::Exception upon any error
   */
  Dv::Xml::Document& parse(const char* data, size_t size) throw (Dv::Xml::Exception) ;

  /** Parse a Dv::Xml::Document from a filenm. The
   * result replaces the current contents of this document.
   * @param filenm name of file where document can be found.
   * @return reference to this document
   * @exception Dv::Xml::Exception upon any error
   */
  Dv::Xml::Document& parse(const std::string& filenm) throw (Dv::Xml::Exception) ;

  /** Set root.
   * @param ref reference to node to copy to root of this document.
   * @exception Dv::Xml::Exception if ref does not refer to a valid xml::node.
   * @sa Dv::Xml::Document::root(const Dv::Xml::Node& node)
   */
  Document& operator=(const Dv::Xml::Node::Ref& ref) throw (Dv::Xml::Exception) ;

  /** Set root.
   * @param ref reference to node that will be copied to the root of this document.
   * @exception Dv::Xml::Exception if node is not valid.
   * @sa Dv::Xml::Node
   */
  void root(const Dv::Xml::Node::Ref& ref) throw (Dv::Xml::Exception) ;

  /** Get root Dv::Xml::Node of Document.
   * @return root Dv::Xml::Node of Document.
   * @warning Assigning the result does \b not change the root.
   * Updating is possible, as in the example below.
   * \code
   * Dv::Xml::Document doc;
   * ..
   * (doc.root() >> "child" >> "grandchild")["name"] = "Pete";
   * \endcode
   */
  Dv::Xml::Node::Ref root() const;

  /** Convert to reference to root of document.
   * @sa Dv::Xml::Document::root
   */
  operator Dv::Xml::Node::Ref() const { return root(); }

  /** Get encoding of Document. Default is "ISO-8859-1". 
   * @return encoding of Document.
   */
  const std::string& encoding() const { return get_encoding(); }

  /** Set encoding of Document. 
   * @param enc encoding of Document
   * @return *this
   */
  Dv::Xml::Document& encoding(const std::string& enc) { 
    set_encoding(enc.c_str()); return *this; }

  /** Validate document w.r.t. DTD.
   * @param dtd url or filename of DTD, if empty, the dtd is
   *   assumed to have been included in the document.
   * @return true iff the document is valid w.r.t the DTD.
   */
  bool valid(const std::string& dtd = "");

  /** Select node after path from document.
   * @param path of the form "a/b/c"
   * @return node at the end of the path
   * @exception Dv::Xml::Exception if node is not valid.
   */
  Dv::Xml::Node::Ref select(const std::string& path) const throw (Dv::Xml::Exception) ;
};

}}

#endif
